package Censor;
require Exporter;
use strict;
use Bio::SearchIO;
use lib ".";
use HashTree;
use vars (
   qw(@ISA @EXPORT $VERSION $EXEC $CPUS $QUMNSZ $QUNSZ $QUMXSZ $DBMNSZ $DBNSZ $DBMXSZ $DEBUG $LOGFILE
     $LOG $SPEC $TMP $DBT $QUT $MPT $DBX $QUX $DBCM1 $QUCM1 $DBCM2 $QUCM2 $DBLL $RALN $LCLS $BOT $BLASTENGINE
     %LIBMAP %DATA %BPRG %PROG %FILT %MODE %MATRIX %BPRM %FILTER %BPARAMS %SUPPORTS %PARAMS),
   qw($DBSZ $DBGC $BP $DB $QU $MT $FT $MP $MD)
);
@ISA    = qw(Exporter);
@EXPORT = qw(
  $VERSION $EXEC $CPUS $QUMNSZ $QUNSZ $QUMXSZ $DBMNSZ $DBNSZ $DBMXSZ $DEBUG $LOGFILE $LOG $SPEC $TMP
  $DBT $QUT $MPT $DBX $QUX $DBCM1 $QUCM1 $DBCM2 $QUCM2 $DBLL $RALN $LCLS $BOT $BLASTENGINE
  %LIBMAP %DATA %BPRG %PROG %FILT %MODE %MATRIX %BPRM %FILTER %BPARAMS %SUPPORTS %PARAMS
  searchLibraries getCPUS getParams Usage runCensX checkParams checkPROG findLib printLog
);

#our @EXPORT_OK = qw();
$SPEC        = "";
$BLASTENGINE = "ncbi-blast";

#$BLASTENGINE = "wu-blast";
$DBSZ = 0;    #Effective size of all database sequences
$DBGC = 0;    #GC content of all database sequences

#my ( %FAMTREE_LIST, %FAMTREE_HASH );
$TMP      = "tmp.$$.tmp";
$DBT      = "db.$$.tmp";       #DB temp file
$QUT      = "qu.$$.tmp";       #QU temp file
$MPT      = "mp.$$.tmp";       #MP temp file
$DBX      = "dbmx.$$.tmp";     #DB reformated file
$QUX      = "qumx.$$.tmp";     #QU reformated file
$DBCM1    = "dbcm1.$$.tmp";    #DB conversion map for ssplit
$QUCM1    = "qucm1.$$.tmp";    #QU conversion map for ssplit
$DBCM2    = "dbcm2.$$.tmp";    #DB conversion map for sjoin
$QUCM2    = "qucm2.$$.tmp";    #QU conversion map for sjoin
$DBLL     = "dbll.$$.tmp";     #DB locus list
$RALN     = "raln.$$.tmp";     #Temporary DB file for realignment
$LCLS     = "lcls.$$.tmp";
$BOT      = "bout.$$.tmp";     #Blast report output
%LIBMAP   = ();
%DATA     = ();
%BPRG     = ();
%PROG     = ();
%FILT     = ();
%MODE     = ();
%MATRIX   = ();
%FILTER   = ();
%BPARAMS  = ();
%SUPPORTS = ();
%PARAMS   = ();
my $code;

#==================================================
sub getParams {
   my $PARAMS = shift;
   if ( getOptions( $PARAMS, \@_ ) != 0 ) {
      return 10;
   }
   $PARAMS->{db} = shift;
   if ( defined( $PARAMS->{db} ) ) {
      if ( -e $PARAMS->{db} && !-z $PARAMS->{db} ) {
         my $CMD = "$EXEC $PARAMS->{db}";
         foreach my $k ( keys %{$PARAMS} ) {
            if ( ref( $PARAMS->{$k} ) =~ /ARRAY/ ) {
               $CMD .= " -$k '" . join( " ", @{ $PARAMS->{$k} } ) . "'";
            }
            else {
               $CMD .= " -$k '$PARAMS->{$k}'";
            }
         }
         printLog( $CMD, $LOG );
         return 0;
      }
      else {
         warn "FATAL: Bad database $PARAMS->{db}: file is empty or does not exist";
         printLog( "FATAL: Bad database $PARAMS->{db}: file is empty or does not exist", $LOG );
      }
   }
   return 20;
}

#==================================================
sub getOptions {
   my $PARAMS = shift;
   my $ARGV   = shift;
   for ( my $i = 0 ; $i <= $#{$ARGV} ; $i++ ) {
      if ( $ARGV->[$i] =~ /^-\S+/ ) {
         if ( $ARGV->[$i] =~ /=/ ) {
            if ( $ARGV->[$i] =~ s/^(\S+)=(\S+)// ) {
               splice @{$ARGV}, $i, 1, ( $1, $2 );
            }
            if ( $ARGV->[$i] =~ s/^(\S+)=/=/ ) {
               splice @{$ARGV}, $i, 1, ($1);
            }
            if ( $ARGV->[$i] =~ s/^=(\S+)/=/ ) {
               splice @{$ARGV}, $i, 1, ($1);
            }
            next;
         }
         $ARGV->[$i] =~ s/^-+//;
         if ( defined $SUPPORTS{PRESETS}{ $ARGV->[$i] } ) {
            if ( ref( $SUPPORTS{PRESETS}{ $ARGV->[$i] }[0] ) =~ /ARRAY/ ) {
               splice @{$ARGV}, $i, 1, @{ $SUPPORTS{PRESETS}{ $ARGV->[$i] }[0] };
               $i--;
               next;
            }
            else {
               die "FATAL: Expecting to have array reference as preset value. Got ", ref $SUPPORTS{PRESETS}{ $ARGV->[$i] }[0];
            }
         }
         if ( defined $SUPPORTS{FLAGS}{ $ARGV->[$i] } ) {
            $PARAMS{ $ARGV->[$i] } = 1;
            splice @{$ARGV}, $i, 1;
            $i--;
            next;
         }
         if ( defined $SUPPORTS{OPTIONS}{ $ARGV->[$i] } ) {
            my ( $key, $value ) = splice @{$ARGV}, $i, 2;
            if ( $key eq "lib" ) {
               foreach my $lib ( split /\s+|,\s*/, $value ) {
                  push @{ $PARAMS->{$key} }, ( defined $LIBMAP{$lib} ) ? $LIBMAP{$lib} : $lib;
               }
            }
            else {
               if ( $SUPPORTS{OPTIONS}{$key}[2] == 1 ) {
                  $PARAMS->{$key} .= " $value";
               }
               elsif ( $SUPPORTS{OPTIONS}{$key}[2] == 2 ) {
                  push @{ $PARAMS->{$key} }, $value;
               }
               else {
                  $PARAMS->{$key} = $value;
               }
            }
            $i--;
            next;
         }
         warn "Unrecognized option '$ARGV->[$i]'\n";
      }
   }
   return 0;
}

#==================================================
sub runCensX {
   my $LIB    = $PARAMS{lib};
   my $db     = convertToFASTA( $PARAMS{db} );
   my $map    = $PARAMS{map};
   my $masked = $PARAMS{masked};
   my $found  = $PARAMS{found};
   my $tab    = $PARAMS{tab};
   my %dbstat;
   my %qustat;
   my $CMD;
   my $code;
   $DB = $db;
   $MP = $MPT;

   if ( !defined $PARAMS{classifyonly} ) {
      if ( -e $map ) {
         unlink $map;
      }
      print STDERR "Searching...";
      if ( -e "$DB.dbmx" && -e "$DB.dbcm1" && creationTime($DB) <= creationTime("$DB.dbmx") ) {
         $DBX   = "$DB.dbmx";
         $DBCM1 = "$DB.dbcm1";
      }
      else {
         execute( "$PROG{split} $DB $DBCM1 $DBNSZ $DBMXSZ $DBX", 1 );
      }
      if ( -e $DBX && !-z $DBX ) {
         $DB = $DBX;
         undef %{ $DATA{$DB} };
         if ( -e $MP ) {
            printLog( "unlink $MP", $LOG );
            unlink "$MP";
         }
         if ( $code = checkDB() != 0 ) {
            print STDERR "Bad database \"$DB\". Code $code.\n";
            return 20;
         }
         foreach my $QUNAME (@$LIB) {
            $QU = $QUNAME;
            if ( $code = checkQU() != 0 ) {
               print STDERR "Bad query \"$QU\". Code $code.\n";
               return 30;
            }
            getStat( $QU, \%qustat );
            if ( defined $BPARAMS{ getBProg() }{$QUNAME} ) {
               $MD = $QUNAME;
            }
            else {
               $MD = $MODE{ $PARAMS{mode} };
            }
            execute( "cat $QU > $QUX", 1 );
            $QU = $QUX;
            if ( $code = checkQU() != 0 ) {
               print STDERR "Bad query \"$QU\". Code $code.\n";
               return 50;
            }
            undef %{ $DATA{$QU} };
            my $ICNT = 0;
            do {    #This loop will work untill no matches in the current chunk of the database found.
               if ( -e $MP && !-z $MP ) {
                  execute( "$PROG{masker} $MP $DB $TMP -c " . getDBMask(), 1 );
                  execute( "mv $TMP $DB",                                  1 );
                  undef %{ $DATA{$DB} };
                  if ( $code = checkDB() != 0 ) {
                     print STDERR "Bad database \"$DB\". Code $code.\n";
                     return 20;
                  }
               }
               $ICNT++;
               printLog( "$QUNAME \#$ICNT", $LOG );

               #----------------------------------------------------------------
               $code = runBlastEngine();
               if ($code) {
                  printLog( "runBlastEngine returned code $code", $LOG );
                  print STDERR "runBlastEngine returned code $code.\n";
                  return 55;
               }

               #----------------------------------------------------------------
               if ( -e $MP && !-z $MP ) {
                  execute( "cat $MP >> $map", 1 );
                  if ( !-w $map ) {
                     die "\n\nFATAL: Can't write to $map\n\n";
                  }
               }
               else {
                  printLog( "$MP is empty", $LOG );
               }
            } while ( $ICNT < 100 && $MD =~ /smprep.ref/ && -e $MP && !-z $MP );
         }
      }
      close DBLL;
      if ( -e $map && !-z $map ) {
         execute( "$PROG{dmx1} $map $DBCM1 $MP", 1 );
         $DB = $db;
         if (    getDBType() eq 'n'
              && !( $MD eq 'norm' || $MD eq 'sens' )
              && !defined( $PARAMS{nocleansimple} ) )
         {
            cleanSimple($MP);
         }
         if ( !-z $MP ) {
            if ( defined $PARAMS{redundant} ) {
               if ( defined $PARAMS{noself} ) {
                  execute( "$PROG{postpr} $MP -dbt " . getDBType() . " -qut " . getQUType() . " -noselfonly $map", 1 );
               }
               else {
                  execute( "cp $MP $map", 1 );
               }
            }
            else {
               execute( "$PROG{postpr} $MP -dbt " . getDBType() . " -qut " . getQUType() . " -nc " . ( ( defined $PARAMS{noself} ) ? "-noself " : "" ) . " $map", 1 );
            }
         }
         else {
            execute( "mv $MP $map", 1 );
         }
      }
      print STDERR "done\n";
   }
   if ( -e $map && !-z $map ) {
      if ( defined $PARAMS{classify} ) {
         print STDERR "Running classification...";
         $MP = $map;
         $DB = $db;
         if ( $code = classify() != 0 ) {
            print STDERR "classify() failed. Exit code $code\n";
            return 90;
         }
         print STDERR "done\n";
         if ( defined( $PARAMS{classifyonly} ) ) {
            return 0;
         }
      }
      if ( !defined $PARAMS{redundant} ) {
         execute( "$PROG{postpr} -dbt " . getDBType() . " -qut " . getQUType() . " $map $map", 1 );
      }
      if ( !%dbstat ) {
         $DBSZ = getStat( $db, \%dbstat );
         if ( defined $PARAMS{classify} ) {
            foreach my $fml (<$ENV{CENSOR_DATADIR}/*.fml>) {
               my $fmt = $fml;
               $fmt =~ s/\.fml$/.fmt/;
               my $supfam = $fml;
               $supfam =~ s/.*\///g;
               $supfam =~ s/\.fml$//;
               if (0) {
                  $CMD = "$PROG{typedef} $fml|";
                  if ( open( FCMD, $CMD ) ) {
                     while (<FCMD>) {
                        my @R = split /\s+/;
                        $qustat{ $R[0] } = $qustat{$supfam};
                     }
                  }
               }
               else {
                  getStat( $fml, \%qustat );
               }
               if ( open( FCMD, $fmt ) ) {
                  while (<FCMD>) {
                     if ( !/^#/ ) {
                        my @R = split /\s+/;
                        if ( @R >= 2 ) {
                           if ( !defined( $qustat{ $R[0] } ) ) {
                              $qustat{ $R[0] } = $qustat{$supfam};
                           }
                           if ( !defined( $qustat{ $R[1] } ) ) {
                              $qustat{ $R[1] } = $qustat{$supfam};
                           }
                        }
                     }
                  }
               }
            }
         }
      }
      my %IDCOUNT;
      if ( defined $PARAMS{rmmap} ) {
         if ( combineMap( $map, $MPT, \%IDCOUNT ) ) {

            #            sumTab( $MPT, $DBSZ, $tab, $map, \%IDCOUNT );
         }
         else {
            print STDERR "combineMap($map, $MPT) failed\n";
            printLog( "$map was not found.", $LOG );
         }
      }
      else {

         #         $code = sumTab( $map, $DBSZ, $tab );
         #         if ($code) {
         #            print STDERR "sumTab() returned $code\n";
         #         }
      }
      if ( !defined $PARAMS{aln} ) {
         if ( ( $PARAMS{mode} eq 'sens' || $PARAMS{mode} eq 'norm' )
              && getBProg() eq $BPRG{'nn'} )
         {
            $PARAMS{aln} = $PARAMS{db} . ".aln";
            $PARAMS{aln} =~ s/.*\///g;
         }
      }
      if ( defined $PARAMS{aln} ) {
         if ( -e $QUT ) {
            unlink $QUT;
         }
         foreach my $QUNAME ( @{ $PARAMS{lib} } ) {
            if ( ( $QUNAME ne "smprep.ref" ) && findLib( \$QUNAME ) ) {
               execute( "cat $QUNAME >> $QUT", 1 );
            }
            else {
               printLog( "$QUNAME will be ignored in realignment.", $LOG );
            }
         }
         my $FALN;
         open( $FALN, ">$PARAMS{aln}" ) or return 110;
         print STDERR "Running realignment...";
         if ( $code = realign( $map, $db, $QUT, $FALN ) ) {
            print STDERR "failed, code $code\n";
         }
         else {
            print STDERR "done\n";
         }
         close $FALN;
      }
      if ( !-z $map ) {
         if ( ( $code = addCoverage( $map, $MPT, \%dbstat, \%qustat ) ) ) {
            warn "addCoverage($map, $MPT) failed";
            printLog( "addCoverage($map, $MPT) failed.", $LOG );
         }
         if ( !defined $PARAMS{nomasked} ) {
            execute( "$PROG{masker} $map $db $masked -c " . getDBMask() );
         }
         if ( !defined $PARAMS{nofound} ) {
            my ( $FMAP, $FCMD );
            open $FMAP, "$map" or return 135;
            $CMD = "|($PROG{extract} - $db $found) $LOG";
            printLog( $CMD, $LOG );
            open $FCMD, "$CMD" or return 137;
            while (<$FMAP>) {
               my @R = split /\s+/;
               if ( $R[6] eq 'c' ) {
                  print $FCMD "$R[0] $R[2] $R[1]\n";
               }
               else {
                  print $FCMD "$R[0] $R[1] $R[2]\n";
               }
            }
            close $FCMD or print STDERR "$CMD failed\.";
            close $FMAP;
         }
      }
   }
   else {
      print STDERR "$map was not found\n";
      printLog( "$map was not found.", $LOG );
   }
   return 0;
}

#==================================================
sub findFile {
   my ( $file, @path ) = @_;
   foreach my $dir (@path) {
      if ( -e "$dir/$file" ) {
         return "$dir/$file";
      }
   }
   die "FATAL: $file not found in " . join( ':', @path );
}

#==================================================
sub classifyFamilyHashTree {
   my ( $superfamily, $tree, $map, $famseqs ) = @_;
   if ( !defined $tree || !%$tree ) {
      return;
   }
   my @families = keys %$tree;
   if (@families) {
      alignClassify( $superfamily, \@families, $map, $famseqs );
      foreach my $family (@families) {
         classifyFamilyHashTree( $family, new HashTree( $tree->{$family} ), $map, $famseqs );
      }
   }
}

#==================================================
sub classifyFamily {
   my ( $family, $map ) = @_;
   my $famtree = findFile( $family . ".fmt", '.', $ENV{CENSOR_DATADIR} );
   my $famseqs = findFile( $family . ".fml", '.', $ENV{CENSOR_DATADIR} );
   if ( !defined $famtree || !defined $famseqs ) {
      return 0;
   }
   open FIN, $famtree or die "FATAL: Failed to read $famtree";
   my $tree = new HashTree;
   while (<FIN>) {
      s/#.*//;
      my @R = split /\s+/;
      if ( @R >= 2 ) {
         $tree->insertTreeNode( $R[0], $R[1] );
      }
   }

   #$tree->printTree( \*STDOUT );
   if ( defined $tree->{$family} ) {
      push @{ $PARAMS{lib} }, $famseqs;
      return classifyFamilyHashTree( $family, new HashTree( $tree->{$family} ), $map, $famseqs );
   }
}

#==================================================
sub classify {
   my @families = split /\|/, $PARAMS{classify};
   foreach my $family (@families) {
      classifyFamily( $family, $MP );
   }
}

#==================================================
sub alignClassify {
   my ( $superfamily, $familiesref, $map, $famseqs ) = @_;
   if ( @$familiesref == 0 ) {
      return;
   }
   my $LCLS    = "lcls.$$.tmp";
   my $maptmp1 = "mpt1.$$.tmp";
   my $maptmp2 = "mpt2.$$.tmp";
   my $maptmp3 = "mpt3.$$.tmp";
   my ( $CMD, $code );
   my %UMAPHASH = ();
   $MD = "classify";

   if ( !-e $map ) {
      print STDERR "$map not found.\n";
      return 10;
   }

   $CMD = "|$PROG{pullnm} - $famseqs | tr '\t' ' ' > $LCLS";
   printLog( $CMD, $LOG );
   open FCMD, "$CMD" or die "FATAL: Failed to open $CMD";
   print FCMD join( "\n", @$familiesref, "" );
   close FCMD or die "ERROR: Executing $CMD - returned $?";

   if ( grepByCol( $map, 4, $superfamily, $maptmp1 ) ) {
      execute("$PROG{extract} -u $maptmp1 $DB $DBX") or return 10;
      open FMAP, $maptmp1 or return 40;
      my %UCOUNT = ();
      while (<FMAP>) {
         s/^\s+//;
         my @R = split /\s+/;
         $UCOUNT{ $R[0] }++;
         @{ $UMAPHASH{ $R[0] . "_" . $UCOUNT{ $R[0] } } } = @R;
      }
      close FMAP;
      my $DBSAVE = $DB;
      $DB = $DBX;
      $QU = $LCLS;
      $MP = $maptmp3;
      undef %{ $DATA{$QU} };
      undef %{ $DATA{$DB} };

      if ( $code = checkQU() != 0 ) {
         print STDERR "Bad query \"$QU\". Code $code.\n";
         return 80;
      }
      if ( $code = checkDB() != 0 ) {
         print STDERR "Bad database \"$DB\". Code $code.\n";
         return 90;
      }

      #----------------------------------------------------------------
      $code = runBlastEngine();
      if ($code) {
         printLog( "runBlastEngine returned code $code", $LOG );
         print STDERR "runBlastEngine returned code $code.\n";
         return 55;
      }
      if ( -z $MP ) {
         warn "\nNo $superfamily members found on this stage.";
      }
      $DB = $DBSAVE;

      #----------------------------------------------------------------
      if ( !-z $MP ) {
         my $CLSMAPREF = bestHits( $superfamily, $familiesref, $MP );
         open FMAP, ">$maptmp3" or return 35;
         foreach my $k ( keys %$CLSMAPREF ) {
            if ( defined $UMAPHASH{$k} && @{ $UMAPHASH{$k} } ) {
               $UMAPHASH{$k}[3] = $CLSMAPREF->{$k}[3];
               $UMAPHASH{$k}[4] = $CLSMAPREF->{$k}[4];
               $UMAPHASH{$k}[5] = $CLSMAPREF->{$k}[5];
               $UMAPHASH{$k}[7] = $CLSMAPREF->{$k}[7];
               $UMAPHASH{$k}[8] = $CLSMAPREF->{$k}[8];
               $UMAPHASH{$k}[9] = $CLSMAPREF->{$k}[9];
               print FMAP ( join "\t", @{ $UMAPHASH{$k} } ), "\n";
            }
            else {
               print STDERR "$k was not found.\n";
            }
         }
         close FMAP;
         $code = ungrepByCol( $map, 4, $superfamily, $maptmp2 );

         execute("cat $maptmp2 $maptmp3 > $maptmp1") or return 20;
         if ( defined $PARAMS{redundant} ) {
            execute("cp $maptmp1 $map") or return 30;
         }
         else {
            execute( "$PROG{postpr} $maptmp1 -nc -dbt " . getDBType() . " -qut " . getQUType() . " $map" ) or return 40;
         }
      }
   }
   unlink $maptmp1;
   unlink $maptmp2;
   unlink $maptmp3;
   return 0;
}

#==================================================
sub bestHits {
   my ( $superfamily, $familiesref, $mapfile ) = @_;
   open FCLASSMAP, $mapfile or die "ERROR: Failed to read $mapfile";
   my %MAPHASH;
   my $delta = 0.005;
   while (<FCLASSMAP>) {
      chomp;
      my @R = split /\s+/;
      if ( defined $MAPHASH{ $R[0] } ) {
         if ( abs( $MAPHASH{ $R[0] }[9] - $R[9] ) / $MAPHASH{ $R[0] }[9] < $delta ) {
            if ( $MAPHASH{ $R[0] }[9] >= $R[9] ) {
               $MAPHASH{ $R[0] }[3] = $superfamily;
               next;
            }
            $R[3] = $superfamily;
         }
         if ( $MAPHASH{ $R[0] }[9] > $R[9] ) {
            next;
         }
      }
      @{ $MAPHASH{ $R[0] } } = @R;
   }
   close FCLASSMAP;
   return \%MAPHASH;
}

#==================================================
sub cleanSimple {
   my $map = shift;
   my $maptmp = ( $map eq "mpt1.$$.tmp" ) ? "mpt2.$$.tmp" : "mpt1.$$.tmp";

   #   my $CMD    = "$PROG{extract} $map $DB | $PROG{dust} - | typedef - |";
   my $CMD = "$PROG{extract} $map $DB | $PROG{nseg} - 21 1.2 1.3 -x | typedef - |";
   printLog( $CMD, $LOG );
   open FCMD,  $CMD       or print STDERR "Failed to open $CMD\n"     and return 10;
   open FIMAP, $map       or print STDERR "Failed to read $map\n"     and return 20;
   open FOMAP, ">$maptmp" or print STDERR "Failed to write $maptmp\n" and return 30;
   my ( $tdline, $mpline );
   while ( defined( $tdline = <FCMD> ) and defined( $mpline = <FIMAP> ) ) {
      my @RTD = split /\s+/, $tdline;
      my @RMP = split /\s+/, $mpline;
      if ( $RMP[3] !~ /\)n$/ ) {

         #The following expression is for estimating signifficance of the remaining
         #unmasked portion of potential false positive:
         # masked_len * sim * masked_len / unmasked_len < 25
         if ( $RTD[2] == 0 ) {
            next;
         }
         my $SIM = ( $RMP[7] * $RTD[3] - ( $RTD[3] - $RTD[2] ) * 0.99 ) / $RTD[2];
         if ( $RTD[2] * $SIM * $RTD[2] / $RTD[3] < 25 ) {
            next;
         }
      }
      print FOMAP join( "\t", @RMP ), "\n";
   }
   close FOMAP;
   close FIMAP;
   close FCMD;
   execute( "mv $maptmp $map", 1 );
}

#==================================================
sub realign {
   my ( $MAP, $FDB, $FQU, $ALN ) = @_;
   my ( $MAPH, $PIPE, $RDDBH, $WRDBH );
   my $maptmp;
   my $FOMAP;
   if ( getBProg() eq $BPRG{'nn'} && !defined( $PARAMS{nocleanfp} ) ) {
      $maptmp = ( $MAP eq "mpt1.$$.tmp" ) ? "mpt2.$$.tmp" : "mpt1.$$.tmp";
      open $FOMAP, ">$maptmp"
        or print STDERR "Failed to write $maptmp\n" and return 30;
   }
   my $QUNAME;
   my %MAPHASH;
   open( $MAPH, $MAP ) or return 10;
   while (<$MAPH>) {
      s/^\s+//;
      my @R = split /\s+/;
      $QUNAME = $R[3];
      if ( !defined $MAPHASH{$QUNAME} ) {
         if ( $QUNAME =~ /^\([ATCG]+\)n$/i ) {
            if ( defined $FOMAP ) {
               print $FOMAP $_;
            }
            next;
         }
         if ( !execute("echo '$QUNAME' | $PROG{pullnm} - $FQU $QUX") ) {
            $MAPHASH{$QUNAME} = 1;
            printLog( "WARNING: Failed to extract '$QUNAME' from '$FQU'", $LOG );
            next;
         }
         my $CMD = "|$PROG{extract} - -u $FDB $DBT $LOG";
         printLog( $CMD, $LOG );
         open( $PIPE, $CMD ) or return 20;
         my $FPOS = tell $MAPH;
         do {
            chomp;
            s/^\s+//;
            @R = split /\s+/;
            if ( $QUNAME eq $R[3] ) {
               #if($R[6] eq 'c') {
               #   print $PIPE join("\t", $R[0], $R[2], $R[1]), "\n";
               #}
               #else
               {
                  print $PIPE join( "\t", $R[0], $R[1], $R[2] ), "\n";
               }
               push @{ $MAPHASH{ $R[3] }{ $R[0] } }, $_;
            }
         } while (<$MAPH>);
         seek $MAPH, $FPOS, 0;
         close $PIPE;

         $DB = $DBT;
         $QU = $QUX;
         $MP = $MPT;
         $MD = "realign";
         if ( $code = checkQU() != 0 ) {
            print STDERR "Bad query \"$QU\". Code $code.\n";
            return 80;
         }
         if ( $code = checkDB() != 0 ) {
            print STDERR "Bad database \"$DB\". Code $code.\n";
            return 90;
         }

         #========================================================
         undef %{ $DATA{$QU} };
         undef %{ $DATA{$DB} };
         getDBType();
         getQUType();
         getBProg();
         getMatrix();
         getFilter();

         if ( $BLASTENGINE eq "wu-blast" ) {
            $CMD = "$BP $DB $QU $BPARAMS{$BP}{$MD} $MT $FT $PARAMS{bprm} 2>&1|";
         }
         else {
            $CMD = "$BPARAMS{$BP}{$MD} -d $DB -i $QU $MT $FT $PARAMS{bprm} 2>&1|";
         }
         printLog( $CMD, $LOG );
         open $PIPE, "$CMD" or return 140;

         #========================================================
         #my $report = new Bio::Tools::BPlite(-fh=>$PIPE);
         my $report = new Bio::SearchIO( -format => 'blast', -fh => $PIPE );
         while ( my $result = $report->next_result() ) {
            my $qun = $result->query_name();

            #warn "query= $qun";
            while ( my $hit = $result->next_hit ) {

               #warn "hitname= " . $hit->name;
               my ( $LOC, $IDX ) = $hit->name =~ /^(\S+)_(\d+)$/;
               $IDX--;

               # process the Bio::Search::Hit::HitI object
               while ( my $hsp = $hit->next_hsp ) {
                  if ( defined( $MAPHASH{$qun}{$LOC}[$IDX] )
                       && $MAPHASH{$qun}{$LOC}[$IDX] )
                  {

                     # print $ALN "$MAPHASH{$qun}{$LOC}[$IDX] >\n";
                     my $dbn = $LOC;
                     my ( $qb, $qe, $sb, $se ) = ( $hsp->start('query'), $hsp->end('query'), $hsp->start('subject'), $hsp->end('subject') );
                     my $dir = $hsp->strand('query') * $hsp->strand('subject') > 0
                       || $hsp->strand('query') * $hsp->strand('subject') == 0 && $hsp->strand('query') + $hsp->strand('subject') >= 0 ? 'd' : 'c';

                     #warn "qdir=" . $hsp->strand('query') . " sdir=" . $hsp->strand('subject') . " $dir $qb-$qe $sb-$se\n";
                     @R = split /\s+/, $MAPHASH{$qun}{$LOC}[$IDX];

                     #if ( $dir ne 'd') {
                     if ( $R[6] ne $dir ) {
                        #warn "\nStrands didn't match:\n";
                        #warn "$dbn\t$sb\t$se\t$qun\t$qb\t$qe\t$dir!= $R[0]\t$R[1]\t$R[2]\t$R[3]\t$R[4]\t$R[5]\t$R[6]\n";
                        next;
                     }
                     if ( defined $PARAMS{noself} ) {
                        if (    $LOC eq $qun
                             && $R[1] - 1 + min( $sb, $se ) == min( $qb, $qe )
                             && $R[1] - 1 + max( $sb, $se ) == max( $qb, $qe ) )
                        {
                           next;
                        }
                     }
                     my $sim = ( $hsp->matches )[0] / ( $hsp->length - $hsp->gaps('query') - $hsp->gaps('subject') + gapCount( $hsp->hit_string ) + gapCount( $hsp->query_string ) );
                     my $pos = ( $hsp->matches )[1] / ( $hsp->length - $hsp->gaps('query') - $hsp->gaps('subject') + gapCount( $hsp->hit_string ) + gapCount( $hsp->query_string ) );
                     my $mapline = sprintf "%-12s %7ld %-7ld %12s %7ld %-7ld %s   %-5.4f %-5.4f %d",
                       $dbn, $R[1] - 1 + min( $sb, $se ), $R[1] - 1 + max( $sb, $se ),
                       $qun, min( $qb, $qe ), max( $qb, $qe ),
                       $R[6], $sim, $pos, $hsp->score;

                     #warn "OLD: $MAPHASH{$qun}{$LOC}[$IDX]\n";
                     #warn "NEW: $mapline\n";
                     
                     if ( defined $FOMAP ) {
                        my $mm2tr = mismatch2transition( $hsp->hit_string, $hsp->query_string );
                        if ( $sim * $hsp->length > 100 || $sim >= 0.75 || $mm2tr <= 2 ) {
                           print $FOMAP "$mapline\n";
                        }
                        else {
                           printf(STDERR "\nRejected (MM/TR=%.2f), : $mapline\n", $mm2tr);
                           next;
                        }
                     }
                     printf $ALN "$mapline <\n\n";
                     printAln( $ALN, 70, $hsp->hit_string, $hsp->homology_string, $hsp->query_string );
                     $MAPHASH{$qun}{$LOC}[$IDX] = 0;
                     last;
                  }
               }
            }
         }
         close $PIPE;
      }
   }
   close $MAPH;
   if ( defined $FOMAP ) {
      close $FOMAP;
      execute("sort -k1,1 -k2,2n $maptmp > $MAP");
   }
   return 0;
}

#==================================================
sub getNCount {
   my ( $S, $Q ) = @_;
   my $ls   = length($S);
   my $lq   = length($Q);
   my $ncnt = 0;
   my $chr  = ( getBProg() eq $BPRG{nn} ) ? 'N' : 'X';
   if ( $ls == $lq ) {
      for ( my $i = length($S) - 1 ; $i >= 0 ; $i-- ) {
         my $cs = uc substr( $S, $i, 1 );
         my $cq = uc substr( $Q, $i, 1 );
         if ( ( $cs eq $chr || $cq eq $chr ) && $cs ne '-' && $cq ne '-' ) {
            $ncnt++;
         }
      }
   }
   return $ncnt;
}

#==================================================
sub gapCount {
   my $ALN = shift;
   return ( $ALN =~ s/(\-+)/$1/g ) + 0;
}

#==================================================
sub runBlastEngine {
   my $RES;
   if ( $BLASTENGINE eq "wu-blast" ) {
      $RES = runWUBlast(@_);
   }
   else {
      $RES = runNCBIBlast(@_);
   }
   return $RES;
}

#==================================================
sub runWUBlast {
   getBProg();
   getMatrix();
   getFilter();
   $| = 1;
   my $qufirst = 0;
   my $CMD     = "$BP $DB $QU $BPARAMS{$BP}{$MD} $MT $FT $PARAMS{bprm} 2>> $LOGFILE |";
   if ( $DATA{$QU}{count} / $DATA{$DB}{count} >= 4 && $DATA{$QU}{count} > 8 ) {
      if ( formatDB($QU) != 0 ) {
         return 10;
      }
      getBProg4Type( getSEQType($QU), getSEQType($DB) );
      $CMD     = "$BP $QU $DB $BPARAMS{$BP}{$MD} $MT $FT $PARAMS{bprm} 2>> $LOGFILE |";
      $qufirst = 1;
   }
   printLog( $CMD, $LOG );
   open FPIPE, $CMD   or return 10;
   open FMAP,  ">$MP" or return 20;
   while (<FPIPE>) {
      my @BLAST = split /\s+/;
      if ( @BLAST == 22 ) {
         my @R;
         if ($qufirst) {

            #0    1     2     3  4        5   6          7        8     9     10       11       12    13       14    15       16       17       18    19       20       21
            #qid  sid   E     N  Sprime   S   alignlen   nident   npos  nmism pcident  pcpos    qgaps qgaplen  sgaps sgaplen  qframe   qstart   qend  sframe   sstart   send
            #qu   sub   2e-9  1  315.89   883 162        162      162   0     100.00   100.00   0     0        0     0        +0       1        162   +1       6595     7080

            @R = (
                   @BLAST[ 0, 17, 18, 1, 20, 21, 16, 19, 5 ],
                   $BLAST[7] / ( $BLAST[6] - $BLAST[13] - $BLAST[15] + $BLAST[12] + $BLAST[14] ),
                   $BLAST[8] / ( $BLAST[6] - $BLAST[13] - $BLAST[15] + $BLAST[12] + $BLAST[14] )
            );
         }
         else {
            @R = (
                   @BLAST[ 1, 20, 21, 0, 17, 18, 16, 19, 5 ],
                   $BLAST[7] / ( $BLAST[6] - $BLAST[13] - $BLAST[15] + $BLAST[12] + $BLAST[14] ),
                   $BLAST[8] / ( $BLAST[6] - $BLAST[13] - $BLAST[15] + $BLAST[12] + $BLAST[14] )
            );
         }

         #if ( $R[6] * $R[7] > 0 || $R[6] * $R[7] == 0 && $R[6] + $R[7] >= 0) {
         if ( substr( $BLAST[16], 0, 1 ) eq substr( $BLAST[19], 0, 1 ) ) {
            print FMAP join( "\t", @R[ 0 .. 5 ], 'd', @R[ 9 .. 10, 8 ] ), "\n";
         }
         else {
            print FMAP join( "\t", @R[ 0 .. 5 ], 'c', @R[ 9 .. 10, 8 ] ), "\n";
         }
      }
      else {
         warn "ERROR: Invalid number of columns in blast output: expected 22, got " . scalar(@BLAST);
         warn "       " . $_;
      }
   }
   close FMAP;
   close FPIPE;
   return 0;
}

#==================================================
sub runNCBIBlast {
   getBProg();
   getMatrix();
   getFilter();
   $| = 1;
   if ( !defined $BPARAMS{$BP}{$MD} ) {
      die "FATAL: \$BPARAMS{$BP}{$MD} not defined";
   }
   my $qufirst = 0;
   my $CMD     = "$BPARAMS{$BP}{$MD} -d $DB -i $QU $MT $FT $PARAMS{bprm} 2>> $LOGFILE |";
   if ( $DATA{$QU}{count} / $DATA{$DB}{count} >= 4 && $DATA{$QU}{count} > 8 ) {
      if ( formatDB($QU) != 0 ) {
         return 10;
      }
      getBProg4Type( getSEQType($QU), getSEQType($DB) );
      $CMD     = "$BPARAMS{$BP}{$MD} -d $QU -i $DB $MT $FT $PARAMS{bprm} 2>> $LOGFILE |";
      $qufirst = 1;
   }
   printLog( $CMD, $LOG );
   open FPIPE, $CMD   or return 10;
   open FMAP,  ">$MP" or return 20;
   while (<FPIPE>) {
      my @R = split /\s+/;
      if ( @R == 12 ) {

         #0    1     2        3        4     5     6        7     8        9     10       11
         #Qid  Sid   identity alnlen   miss  gapo  qstart   qend  sstart   send  evalue   bitscore
         #qu   sub   100.00   1621     0     0     16466    11604 16466    11604 0.0      3732
         if ( ( $R[9] - $R[8] ) * ( $R[7] - $R[6] ) >= 0 ) {
            if ($qufirst) {
               print FMAP join( "\t", @R[ 0, 6, 7, 1, 8, 9 ], 'd', $R[2] / 100, $R[2] / 100, $R[11] ), "\n";
            }
            else {
               print FMAP join( "\t", @R[ 1, 8, 9, 0, 6, 7 ], 'd', $R[2] / 100, $R[2] / 100, $R[11] ), "\n";
            }
         }
         else {
            if ($qufirst) {
               print FMAP join( "\t", @R[ 0, 6, 7, 1, 9, 8 ], 'c', $R[2] / 100, $R[2] / 100, $R[11] ), "\n";
            }
            else {
               print FMAP join( "\t", @R[ 1, 9, 8, 0, 6, 7 ], 'c', $R[2] / 100, $R[2] / 100, $R[11] ), "\n";
            }
         }
      }
      else {
         warn "ERROR: Invalid number of columns in blast output: expected 12, got " . scalar(@R);
         warn "       " . $_;
      }
   }
   close FPIPE;
   close FMAP;
   return 0;
}

#==================================================
sub getSEQType {
   my ( $seq, $stathash ) = @_;
   if ( !defined $DATA{$seq}{type} ) {
      my $CMD = "$PROG{typedef} $seq -g |";
      printLog( $CMD, $LOG );
      open( RF, "$CMD" ) or die "FATAL: Failed to open $CMD";
      $DATA{$seq}{count} = 0;
      while (<RF>) {
         my @ls = split /\s+/;
         if ( $ls[0] eq 'Global' ) {
            $DATA{$seq}{type} = $ls[1];
            $DATA{$seq}{size} = $ls[2];
            if ( $DATA{$seq}{type} eq 'n' ) {
               $DATA{$seq}{cg} = $ls[4];
            }
         }
         else {
            $DATA{$seq}{count}++;
            if ( defined $stathash ) {
               @{ $stathash->{ $ls[0] } } = ( $ls[3], $ls[2], $ls[4] );
            }
         }
      }
      close RF;
      if ( $DATA{$seq}{type} eq 'n' ) {
         $DATA{$seq}{mask} = 'X';
      }
      else {
         $DATA{$seq}{mask} = 'X';
      }
   }
   return $DATA{$seq}{type};
}

#==================================================
sub getDBType {
   return getSEQType($DB);
}

#==================================================
sub getQUType {
   return getSEQType($QU);
}

#==================================================
sub getStat {
   my ( $seq, $stathash ) = @_;
   undef $DATA{$seq};
   getSEQType( $seq, $stathash );
   return $DATA{$seq}{size};
}

#==================================================
sub getSEQMask {
   my $seq = shift;
   if ( defined $PARAMS{maskchar} ) {
      return $PARAMS{maskchar};
   }
   if ( !defined $DATA{$seq}{mask} ) {
      if ( getSEQType($seq) eq 'n' ) {
         $DATA{$seq}{mask} = 'X';
      }
      else {
         $DATA{$seq}{mask} = 'X';
      }
   }
   return $DATA{$seq}{mask};
}

#==================================================
sub getDBMask {
   return getSEQMask($DB);
}

#==================================================
sub getQUMask {
   return getSEQMask($QU);
}

#==================================================
sub getBProg4Type {
   my ( $dbt, $qut ) = @_;
   if ( defined( $BP = $PARAMS{bprg} ) ) {
      foreach my $k ( keys %BPRG ) {
         if ( $BP eq $BPRG{$k} && $k =~ /^$dbt$qut/ ) {
            return $BP;
         }
      }
   }
   if ( defined( $BP = $BPRG{ $dbt . $qut } ) ) {
      return $BP;
   }
   die "FATAL ERROR: Can't determin which wu-blast module to use for type combination '$dbt$qut'.\n";
}

#==================================================
sub getBProg {
   my $dbt = getDBType();
   my $qut = getQUType();
   return getBProg4Type( $dbt, $qut );
}

#==================================================
sub getMatrix {
   if ( $PARAMS{bprm} =~ /matrix/ ) {
      $MT = "";
      return $MT;
   }
   my $CG;
   my $MINCG  = 35;    #47; #35;
   my $MAXCG  = 53;    #51; #53;
   my $QUNAME = $QU;
   $QUNAME =~ s/.*\///g;
   if (    defined( $MT = $MATRIX{ getBProg() }{$QUNAME} )
        || defined( $MT = $MATRIX{ getBProg() }{$MD} ) )
   {
      $CG = 0.5;
      if ( defined $DATA{$DB}{cg} ) {
         if ( defined( $DATA{$QU}{cg} ) ) {
            $CG = ( $DATA{$DB}{cg} * $DATA{$DB}{size} + $DATA{$QU}{cg} * $DATA{$QU}{size} ) / ( $DATA{$DB}{size} + $DATA{$QU}{size} );
         }
         else {
            $CG = $DATA{$DB}{cg};
         }
      }
      elsif ( defined( $DATA{$QU}{cg} ) ) {
         $CG = $DATA{$QU}{cg};
      }
      $CG = int( $CG * 50 ) * 2 + 1;
      $CG = $CG < $MINCG ? $MINCG : $CG;
      $CG = $CG > $MAXCG ? $MAXCG : $CG;
      $MT =~ s/<CG>/$CG/;
      return $MT;
   }
   $MT = "";
   return $MT;
}

#==================================================
sub getFilter {
   my $QUNAME = $QU;
   $QUNAME =~ s/.*\///g;
   if ( defined( $FT = $FILTER{$MD} ) ) {
      return $FT;
   }
   if ( $PARAMS{bprm} =~ /filter|wordmask|-F / ) {
      $FT = "";
   }
   else {
      $FT = $FILTER{ getBProg() };
   }

   #   if($MD == $MODE{'norm'} || $MD == $MODE{'sens'}) {
   #      $FT =~ s/wordmask/filter/;
   #   }
   return $FT;
}

#==================================================
sub formatDB {
   my $db = shift;
   if ( $BLASTENGINE eq "wu-blast" ) {
      formatWUBlastDB($db);
   }
   else {
      formatNCBIBlastDB($db);
   }
}

#==================================================
sub formatWUBlastDB {
   my $db = shift;
   my $t  = getSEQType($db);
   if (    $t eq 'n' && -f "$db.xns" && -f "$db.xnd" && -f "$db.xnt"
        || $t eq 'p' && -f "$db.xpd" && -f "$db.xps" && -f "$db.xpt" )
   {
      my @stat1 = stat("$db");
      my @stat2 = stat( $t eq 'n' ? "$db.xns" : "$db.xps" );
      if ( $stat1[9] < $stat2[9] ) {
         return 0;
      }
   }
   execute( "$PROG{xdfmt} -$t -C " . getSEQMask($db) . " $db", 1 );
   return 0;
}

#==================================================
sub formatNCBIBlastDB {
   my $db = shift;
   my $t  = getSEQType($db);
   my $CMD;

   # For some reason formatdb does not digest well tabs in fasta header lines.
   if (    $t eq 'n' && -f "$db.nhr" && -f "$db.nin" && -f "$db.nsq"
        || $t eq 'p' && -f "$db.phr" && -f "$db.pin" && -f "$db.psq" )
   {
      my @stat1 = stat("$db");
      my @stat2 = stat( $t eq 'n' ? "$db.nsq" : "$db.psq" );
      if ( $stat1[9] < $stat2[9] ) {
         return 0;
      }
   }
   if ( $t eq 'n' ) {
      execute( "tr '\t' ' ' < $db | $PROG{formatdb} -i stdin -p F -n $db", 1 );
   }
   else {
      execute( "tr '\t' ' ' < $db | $PROG{formatdb} -i stdin -p T -n $db", 1 );
   }
   return 0;
}

#==================================================
sub checkDB {
   if ( !-e "$DB" ) {
      printLog( "File $DB not found.", $LOG );
      print STDERR "File $DB not found.\n";
      return 10;
   }
   $DB = convertToFASTA($DB);
   if ( $BLASTENGINE eq "wu-blast" ) {
      return formatWUBlastDB($DB);
   }
   else {
      return formatNCBIBlastDB($DB);
   }
}

#==================================================
sub findLib {
   my $FNAME = shift;
   if ( -e "$$FNAME" ) {
      return 1;
   }
   if ( -e "$ENV{REPBASEDIR}/$$FNAME" ) {
      $$FNAME = "$ENV{REPBASEDIR}/$$FNAME";
      return 1;
   }
   printLog( "'$$FNAME' not found in ./ nor in $ENV{REPBASEDIR}/", $LOG );
   print STDERR "'$$FNAME' not found in ./ nor in $ENV{REPBASEDIR}/\n";
   return 0;
}

#==================================================
sub checkQU {
   if ( !findLib( \$QU ) ) {
      return 10;
   }
   $QU = convertToFASTA($QU);
   return 0;
}

#==================================================
# Converts file format to FASTA if neccesary.
# converts line ends to unix format using dos2unix
#==================================================
sub convertToFASTA {
   my $SQ   = shift;
   my $SQFA = "$SQ.fa";
   $SQFA =~ s/.*\///g;
   open( FIN, $SQ ) or die "ERROR: Failed to read $SQ";
   $_ = <FIN>;
   if ( !/^>/ ) {
      if ( execute( "cat $SQ | $PROG{igtofa} - $SQFA", 1 ) ) {
         $SQ = $SQFA;
      }
   }
   else {
      my $cnt = 10;
      while ( defined( $_ = <FIN> ) && --$cnt ) {
         if (/\r/) {
            if ( -w $SQ ) {
               execute( "dos2unix $SQ", 1 );
            }
            else {
               execute( "dos2unix -n $SQ $SQFA", 1 );
               $SQ = $SQFA;
            }
            last;
         }
      }
   }
   close FIN;
   return $SQ;
}

#==================================================
sub execute {
   my ( $CMD, $die_on_error ) = @_;
   system "echo \"$CMD\" >> $LOGFILE";
   $CMD = "($CMD) $LOG";
   if ( ( system $CMD ) == 0 ) {
      return 1;
   }
   if ($die_on_error) {
      die "Failed running $CMD";
   }
   return 0;
}

#==================================================
sub Usage {
   open( FHLP, "| pod2text" );
   print FHLP getUsage(), "\n";
   close FHLP;
}

#==================================================
sub searchLibraries {
   my $LIBHASH = shift;
   my $LIBPATH = shift;
   my $LIBCNT  = 0;
   if ( -d "$LIBPATH" ) {
      my $RES = open( FLS, "ls $LIBPATH/*rep.ref |" );
      while (<FLS>) {
         chomp;
         if ( !-z "$_" ) {
            s/.*\///g;
            my $LIB = $_;
            s/rep.*$//;
            $LIBHASH->{$_} = $LIB;
            $LIBCNT++;
         }
      }
      close(FLS);
   }
   return $LIBCNT;
}

#==================================================
sub makeUniqMap {
   my ( $INPMAP, $OUTMAP ) = @_;
   my ( $FINP, $FOUT );
   open( $FINP, "$INPMAP" )  or return 10;
   open( $FOUT, ">$OUTMAP" ) or return 20;
   my %UNIQ = ();
   while (<$FINP>) {
      s/^\s+//;
      my @SRC = split /\s+/, $_, 2;
      $UNIQ{ $SRC[0] }++;
      s/^\S+/$SRC[0]_$UNIQ{$SRC[0]}/;
      print $FOUT $_;
   }
   close $FINP;
   close $FOUT;
   return 0;
}

#==================================================
sub isTransition {
   my ( $bp1, $bp2 ) = @_;
   if ( uc($bp1) eq 'A' ) { return ( uc($bp2) eq 'G' ); }
   if ( uc($bp1) eq 'G' ) { return ( uc($bp2) eq 'A' ); }
   if ( uc($bp1) eq 'T' ) { return ( uc($bp2) eq 'C' ); }
   if ( uc($bp1) eq 'C' ) { return ( uc($bp2) eq 'T' ); }
}

#==================================================
sub isGap {
   my ( $bp1, $bp2 ) = @_;
   return ( $bp1 eq '-' || $bp2 eq '-' );
}

#==================================================
sub mismatch2transition {
   my ( $SQ1, $SQ2 ) = @_;
   my ( $MM, $TR ) = ( 0, 0 );
   if ( getDBType() eq 'n' ) {
      for ( my $i = length($SQ1) - 1 ; $i >= 0 ; $i-- ) {
         my $C1 = substr( $SQ1, $i, 1 );
         my $C2 = substr( $SQ2, $i, 1 );
         if ( $C1 eq $C2 || isGap( $C1, $C2 ) ) {
            next;
         }
         if ( isTransition( $C1, $C2 ) ) {
            ++$TR;
         }
         ++$MM;
      }
      return $TR > 0 ? $MM / $TR : 99;
   }
   return 0;
}

#==================================================
sub printAln {
   my ( $FOUT, $WIDTH, $SQ1, $ALNSTR, $SQ2 ) = @_;
   my ( $s1, $s2, $s3 );
   if ( getDBType() eq 'n' ) {
      for ( my $i = length($SQ1) - 1 ; $i >= 0 ; $i-- ) {
         my $C1 = substr( $SQ1, $i, 1 );
         my $C2 = substr( $SQ2, $i, 1 );
         if ( substr( $ALNSTR, $i, 1 ) eq ' ' ) {
            if ( getDBType() eq 'n' && isTransition( $C1, $C2 ) ) {
               substr( $ALNSTR, $i, 1, ':' );
            }
            elsif ( isGap( $C1, $C2 ) ) {
               substr( $ALNSTR, $i, 1, '-' );
            }
            else {
               substr( $ALNSTR, $i, 1, ' ' );
            }
         }
      }
   }
   else {
      for ( my $i = length($SQ1) - 1 ; $i >= 0 ; $i-- ) {
         if ( substr( $ALNSTR, $i, 1 ) eq ' ' ) {
            if ( isGap( substr( $SQ1, $i, 1 ), substr( $SQ2, $i, 1 ) ) ) {
               substr( $ALNSTR, $i, 1, '-' );
            }
            else {
               substr( $ALNSTR, $i, 1, ' ' );
            }
         }
      }
   }

   #  while((length($SQ1) > $WIDTH) && (length($ALNSTR) > $WIDTH) && (length($SQ2) > $WIDTH)) {
   while (    ( length($SQ1) > 0 )
           && ( length($ALNSTR) > 0 )
           && ( length($SQ2) > 0 ) )
   {
      $SQ1 =~ s/(.{1,$WIDTH})//;
      print $FOUT "S: $1\n";
      $ALNSTR =~ s/(.{1,$WIDTH})//;
      print $FOUT "   $1\n";
      $SQ2 =~ s/(.{1,$WIDTH})//;
      print $FOUT "Q: $1\n\n";
   }

   #  print $FOUT "$SQ1\n$ALNSTR\n$SQ2\n\n";
}

#==================================================
sub swapMaps {
   my $inmap1 = shift;
   my $inmap2 = shift;
   my $outmap = shift;
   my $CODE   = 0;
   my ( $L1, $L2, @LST1, @LST2, $linecnt );
   if ( open( INMAP1, "$inmap1" ) != 1 ) { return 1; }
   if ( open( INMAP2, "$inmap2" ) != 1 ) { close INMAP1; return 2; }
   if ( open( OUTMAP, ">$outmap" ) != 1 ) {
      close INMAP1;
      close INMAP2;
      return 3;
   }
   while ( ( defined( $L1 = <INMAP1> ) ) && ( defined( $L2 = <INMAP2> ) ) ) {
      $linecnt++;
      @LST1 = split /\s+/, $L1;
      @LST2 = split /\s+/, $L2;
      while ( $LST1[0] ne $LST2[0] ) {
         if ( $LST1[0] lt $LST2[0] ) {
            if ($DEBUG) { print STDERR "1: $L1"; }

            #$L1 =~ s/^\s*(\S+)_\d+\s/$1 /;
            #print OUTMAP $L1;
            unless ( defined( $L1 = <INMAP1> ) ) {
               last;
            }
            @LST1 = split /\s+/, $L1;
         }
         else {
            if ($DEBUG) { print STDERR "2: $L2"; }
            unless ( defined( $L2 = <INMAP2> ) ) {
               last;
            }
            @LST2 = split /\s+/, $L2;
         }
      }
      unless ( $L1 && $L2 ) {
         last;
      }
      if ($DEBUG) { print STDERR "1: $L1"; }
      if ($DEBUG) { print STDERR "2: $L2"; }
      if ($DEBUG) { print STDERR "====================\n"; }
      $LST1[0] =~ s/_\d+$//;

      #      $LST1[1] += ($LST2[1] - 1);
      #      $LST1[2] = $LST1[1] + $LST2[2] - $LST2[1];
      $L2 =~ s/^\S+(\s+)\S+(\s+)\S+(\s+)/$LST1[0]$1$LST1[1]$2$LST1[2]$3/;
      print OUTMAP $L2;
   }
   close INMAP1;
   close INMAP2;
   close OUTMAP;
   return $CODE;
}

#==================================================
sub relink {
   my ( $SRC, $DST ) = @_;
   if ( -e $DST ) {
      unlink $DST;
   }
   if ( symlink $SRC, $DST ) { return 0; }
   return 1;
}

#==================================================
sub getCPUS {
   $CPUS = 1;
   my $uname = `uname`;
   chomp $uname;
   if ( $uname eq "AIX" ) {
      $CPUS = `lsdev -C -c processor -S Available|grep -c .`;
      chomp $CPUS;
      return $CPUS;
   }
   if ( $uname =~ /.*BSD|.*bsd|Darwin/ ) {
      $CPUS = `sysctl -n hw.ncpu`;
      chomp $CPUS;
      return $CPUS;
   }
   if ( $uname eq "HP-UX" ) {
      $CPUS = `ioscan -fnkC processor|grep -c '^processor'`;
      chomp $CPUS;
      return $CPUS;
   }
   if ( $uname =~ /IRIX.*/ ) {    # IRIX, IRIX64
      $CPUS = `hinv -c processor|head -1|awk '{print $1}'`;
      chomp $CPUS;
      return $CPUS;
   }
   if ( $uname eq "Linux" ) {
      if ( -f "/proc/cpuinfo" ) {
         $CPUS = `grep -c '^processor' /proc/cpuinfo`;
         chomp $CPUS;
         return $CPUS;
      }
      else {
         print STDERR "No /proc/cpuinfo found.\n";
      }
   }
   if ( $uname =~ /OSF1|SunOS/ ) {
      $CPUS = `psrinfo|grep -c on-line`;
      chomp $CPUS;
      return $CPUS;
   }
   return $CPUS;
}

#==================================================
sub checkParams {
   if ( !-e $PARAMS{db} || !-r $PARAMS{db} || -z $PARAMS{db} ) {
      print STDERR "ERROR: bad database $PARAMS{db}\n";
      printLog( "ERROR: bad database $PARAMS{db}", $LOG );
      return 0;
   }
   foreach my $QU ( @{ $PARAMS{lib} } ) {
      my $QUF = $QU;
      if ($DEBUG) { print STDERR "Checking $QU ... "; }
      findLib( \$QUF ) or return 0;
      if ($DEBUG) { print STDERR "found as $QUF\n"; }
   }
   return 1;

   #return (    checkPROG( $ENV{PATH}, values %BPRG )
   #         && checkPROG( $ENV{PATH},          values %PROG )
   #         && checkPROG( $ENV{WUBLASTFILTER}, values %FILT ) );
}

#==================================================
sub checkPROG {
   my ( $path, @progs ) = @_;
   my %path_hash = ();
   foreach my $prg (@progs) {
      undef %path_hash;
      if ($DEBUG) { print STDERR "Checking for $prg in ... "; }
      my $found = 0;
      foreach my $dir ( split /:/, $path ) {
         if ( defined $path_hash{$dir} ) {
            next;
         }
         $path_hash{$dir} = 1;
         if ($DEBUG) { print STDERR "\n\t$dir"; }
         if ( -e "$dir/$prg" ) {
            if ($DEBUG) { print STDERR " exists"; }
            if ( -x "$dir/$prg" ) {
               if ($DEBUG) { print STDERR ", executable\n"; }
               $found = 1;
               last;
            }
            else {
               next;
            }
         }
      }
      if ( !$found ) {
         print STDERR "\nERROR: Executable $prg was not found in ", join ":", keys %path_hash, "\n";
         printLog( "ERROR: Executable $prg was not found in $path", $LOG );
         return 0;
      }
   }
   return 1;
}

#==================================================
sub min {
   my ( $a, $b ) = @_;
   return $a < $b ? $a : $b;
}

#==================================================
sub max {
   my ( $a, $b ) = @_;
   return $a > $b ? $a : $b;
}

#==================================================
sub minx {
   @_ == 1 ? return shift : return min( shift, minx(@_) );
}

#==================================================
sub maxx {
   @_ == 1 ? return shift : return max( shift, maxx(@_) );
}

#==================================================
sub addCoverage {
   my ( $map, $MPT, $dbstat, $qustat ) = @_;
   open( FINP, $map )    or warn "Failed to read $map"  and return 10;
   open( FOUT, ">$MPT" ) or warn "Failed to write $MPT" and return 20;
   while (<FINP>) {
      chomp;
      my @R = split /\s+/;
      if ( $dbstat->{ $R[0] } && $qustat->{ $R[3] } ) {
         if (    $dbstat->{ $R[0] }->[0] >= $R[2]
              && $qustat->{ $R[3] } >= $R[5] )
         {
            printf FOUT "$_\t%.2f\t%.2f\n", min( 100, 100 * ( 1 + abs( $R[2] - $R[1] ) ) / $dbstat->{ $R[0] }->[0] ),

              #              min(100, 100 * ( 1 + abs( $R[5] - $R[4] ) ) / $qustat->{ $R[3] }->[0]);
              min( 100, 100 * ( 1 + min( max( $R[5], $R[4] ), $qustat->{ $R[3] }->[0] ) - min( $R[5], $R[4] ) ) / $qustat->{ $R[3] }->[0] );
            next;
         }
      }
      else {
         if ( !defined( $dbstat->{ $R[0] } ) ) {
            warn "Statistics for $R[0] was not found";
         }
         if ( !defined( $qustat->{ $R[3] } ) ) {
            warn "Statistics for $R[3] was not found";
         }
      }
      print FOUT "$_\t-0.00\n";
   }
   close FINP;
   close FOUT;
   execute( "mv $MPT $map", 1 );
   return 0;
}

#==================================================
sub combineMap {
   my ( $input, $output, $IDCOUNT ) = @_;
   my %MAPHASH;
   my ( $INPUT, $OUTPUT, $S, $Q, $ID, $id, @R1, @R2 );
   $ID = 1;
   open $INPUT,  "$input"   or return 0;
   open $OUTPUT, ">$output" or return 0;
   while (<$INPUT>) {
      s/^\s+//;
      chomp;
      @R2 = split /\s+/;
      ( $S, $Q ) = ( $R2[0], $R2[3] );
      splice @R2, 3, 1;
      splice @R2, 0, 1;
      my $NEXT = 1;
      $id = $ID;

      if ( defined @{ $MAPHASH{$S}{$Q} } ) {
         @R1 = @{ $MAPHASH{$S}{$Q} };
         if ( $R1[0] > $R2[0] ) {
            print STDERR "ERROR: Map is not sorted\n";
            return 0;
         }
         if ( $R2[4] eq $R1[4] ) {
            my $SIZE;
            my $SIZE1 = $R1[3] - $R1[2];
            my $SIZE2 = $R2[3] - $R2[2];

            #	if(abs($R2[0] - $R1[1]) <= 2 * ($SIZE1 + $SIZE2))
            {

               # Marking defragmented elements as one
               if ( $R2[4] eq "d" ) {

                  # In direct orientation
                  $SIZE = $R2[3] - $R1[2];

                  #print STDOUT "$SIZE - ($SIZE1 + $SIZE2) = ", abs($SIZE - ($SIZE1 + $SIZE2)), "; 0.15 * min($SIZE1, $SIZE2) = ", 0.15 * min($SIZE1, $SIZE2), "\n";
                  if ( abs( $SIZE - ( $SIZE1 + $SIZE2 ) ) < ( 0.1 * ( $SIZE1 + $SIZE2 ) ) ) {
                     $id   = $R1[$#R1];
                     $NEXT = 0;
                  }
               }
               else {

                  #In oposite orientation
                  $SIZE = $R1[3] - $R2[2];

                  #print STDOUT "$SIZE - ($SIZE1 + $SIZE2) = ", abs($SIZE - ($SIZE1 + $SIZE2)), "; 0.15 * min($SIZE1, $SIZE2) = ", 0.15 * min($SIZE1, $SIZE2), "\n";
                  if ( abs( $SIZE - ( $SIZE1 + $SIZE2 ) ) < ( 0.1 * ( $SIZE1 + $SIZE2 ) ) ) {
                     $id   = $R1[$#R1];
                     $NEXT = 0;
                  }
               }

               #Marking satellites as one element
               #print STDOUT abs($R2[0] - $R1[1]), "; ", abs(2 - ($R2[1] - $R1[0]) / abs(max($R1[3], $R2[3]) - min($R1[2], $R2[2]))), "\n";
               #if(abs($R2[0] - $R1[1]) < 20 && abs(2 - ($R2[1] - $R1[0]) / abs(max($R1[3], $R2[3]) - min($R1[2], $R2[2]))) < 0.2) {
               #$id = $R1[$#R1];
               #$NEXT = 0;
               #}
            }
         }
      }
      if ($NEXT) {
         $ID++;
      }
      @{ $MAPHASH{$S}{$Q} } = ( @R2, $id );
      printf $OUTPUT "$_ %6d\n", $id;
      if ( defined $IDCOUNT ) {
         $IDCOUNT->{$id}++;
      }
   }
   close $INPUT;
   close $OUTPUT;
   return 1;
}

#==================================================
sub recover {
   my $pid    = shift;
   my $mptmp  = "mp.$pid.tmp";
   my $qucm1  = "qucm1.$pid.tmp";
   my $qucm2  = "qucm2.$pid.tmp";
   my $dbcm   = "dbcm1.$pid.tmp";
   my $outmap = "recover.$pid.map";
   if ( -e $mptmp && -e $qucm1 && -e $qucm2 && -e $dbcm ) {
      execute( "$PROG{dmx2} $mptmp $qucm2 | $PROG{dmx2} - $qucm1 | $PROG{dmx1} - dbcm1.17103.tmp | $PROG{postpr} - -dbt n -qut n > $outmap", 1 );
   }
   else {
      warn "ERROR: Missing one or more required files: $mptmp $qucm1 $qucm2 $dbcm";
      return 0;
   }
   return 1;
}

#==================================================
sub grepByCol {
   my ( $input, $CNUM, $WHAT, $output ) = @_;
   my ( $INPUT, $OUTPUT );
   $WHAT = qr /$WHAT/;
   open $INPUT,  $input     or return 0;
   open $OUTPUT, ">$output" or return 0;
   my $CNT = 0;
   while (<$INPUT>) {
      s/^\s+//;
      if (/((\S+)\s+){$CNUM}/) {
         if ( $2 =~ /^$WHAT$/ ) {
            print $OUTPUT $_;
            $CNT++;
         }
      }
   }
   close $INPUT;
   close $OUTPUT;
   return $CNT;
}

#==================================================
sub ungrepByCol {
   my ( $input, $CNUM, $WHAT, $output ) = @_;
   my ( $INPUT, $OUTPUT );
   $WHAT = qr /$WHAT/;
   open $INPUT,  $input     or return 0;
   open $OUTPUT, ">$output" or return 0;
   my $CNT = 0;
   while (<$INPUT>) {
      s/^\s+//;
      if (/((\S+)\s+){$CNUM}/) {
         if ( $2 !~ /^$WHAT$/ ) {
            print $OUTPUT $_;
            $CNT++;
         }
      }
   }
   close $INPUT;
   close $OUTPUT;
   return $CNT;
}

#==================================================
sub printLog {
   my $msg = shift;
   my $log = shift;
   if ($log) {
      system "echo \"$msg\" $log";
   }
   else {
      warn "$msg";
   }
}

#==================================================
sub creationTime {
   my $file = shift;
   my @stat = stat("$file");
   if (@stat) {
      return $stat[9];
   }
   else {
      return undef;
   }
}

#==================================================
sub getUsage {
   my $FLAGLIST = "-" . join " -", sort( keys %{ $SUPPORTS{FLAGS} } );
   my $usage = "
=head1 NAME

$EXEC - compares and masks protein or nucleotide sequences.

=head1 SYNOPSIS

B<$EXEC> database_file
";
   foreach my $k ( sort keys %{ $SUPPORTS{PRESETS} } ) {
      $usage = $usage . "S<[-$k]>\n";
   }
   foreach my $k ( sort keys %{ $SUPPORTS{OPTIONS} } ) {
      $usage = $usage . "S<[-$k " . $SUPPORTS{OPTIONS}{$k}[0] . "]>\n";
   }
   $usage = $usage . "S<[$FLAGLIST]>

=head1 DESCRIPTION

B<$EXEC> aligns and censors similar fragments of two sequences.
At least one input file, referred to in this document as a database, should be provided.
Input files must be in IG or FASTA format.
By default B<$EXEC> removes low-complexity regions from the database,
running it against library of simple repeats, and from the query, using low-complexity filters.

Any combination of protein or nucleotide databse or query is allowed.
B<$EXEC> determines the type of input sequences (nucleotide or amino acid) by calculating density
of 'ATCGN' bases in each sequence. Depending on the combination of types of database and query
sequences B<$EXEC> invokes corresponding blast module with specific parameters.

Output includes:

=over 4

=item 1.

Map of found matches with coordinates of corresponding fragments in database and query. If no matches were found no map will be produced.\n

=item 2.

Masked database sequences. Masking means the replacement of matched fragments with the masking symbol 'N'('X').\n

=item 3.

Fragments extracted from the database that correspond to the map.

=item 4.

Optional local alignments of matched fragments.

=back

The full version of B<$EXEC> goes with a copy of B<Repbase> - database of repetitive elements for different species,
that is used as a reference in detection of repeats. The latest full version of the package or Repbase separately
can be downloaded from the website of B<Genetic Information Research Institute>:

B<http://www.girinst.org/server/censor/>

B<http://orka.girinst.org/server/RepBase/>

Registration is required, and it is free for nonprofit research arganizations.

$SPEC

=head1 ARGUMENTS

=head2 database

Contains target sequences.

=head2 query

Contains reference sequences.

For example, if you need to run human chromosome against library of human repeats,
then the chromosome sequence file should be specified as database, and library of repeats as query, using -lib option.

=head1 PRESETS OF OPTIONS
";
   foreach my $k ( sort keys %{ $SUPPORTS{PRESETS} } ) {
      $usage = $usage . "\n=head2 -$k\n\nequals \"" . join( " ", @{ $SUPPORTS{PRESETS}{$k}[0] } ) . "\"\n\n$SUPPORTS{PRESETS}{$k}[1]\n";
   }
   $usage = $usage . "
=head1 OPTIONS
";
   foreach my $k ( sort keys %{ $SUPPORTS{OPTIONS} } ) {
      $usage = $usage . "\n=head2 -$k $SUPPORTS{OPTIONS}{$k}[0]\n\n$SUPPORTS{OPTIONS}{$k}[1]\n";
   }
   $usage = $usage . "
=head1 FLAGS
";

   #while(my @ls = each %{$SUPPORTS{FLAGS}}) {
   #  $usage = $usage . "\n=head2 -$ls[0]\n\n$ls[1]\n";
   #}
   foreach my $k ( sort keys %{ $SUPPORTS{FLAGS} } ) {
      $usage = $usage . "\n=head2 -$k\n\n$SUPPORTS{FLAGS}{$k}\n";
   }
   $usage = $usage . "
=head1 DEFAULTS

Default command line parameters for each blast program for different sensitivity levels are listed below.

";
   foreach my $bp ( sort values %BPRG ) {
      foreach my $md ( sort keys %MODE ) {
         if ( $md ne 'none' ) {
            if ( $BLASTENGINE eq "wu-blast" ) {
               $usage =
                   $usage 
                 . "I<$md:> B<$bp> DB QU $BPARAMS{$bp}{$MODE{$md}} $MATRIX{$bp}{$MODE{$md}} "
                 . (
                     defined( $FILTER{ $MODE{$md} } )
                     ? $FILTER{ $MODE{$md} }
                     : $FILTER{$bp}
                 ) . "\n\n";
            }
            else {
               $usage =
                   $usage 
                 . "I<$md:> $BPARAMS{$bp}{$MODE{$md}} $MATRIX{$bp}{$MODE{$md}} -i QU -d DB "
                 . (
                     defined( $FILTER{ $MODE{$md} } )
                     ? $FILTER{ $MODE{$md} }
                     : $FILTER{$bp}
                 ) . "\n\n";
            }
         }
      }
   }
   $usage = $usage . "
=head1 TIPS AND HINTS

=head2 FILTER

If you don't need blast's filter module to be executed on query sequences, add  '-filter=none' to -bprm option line.

e.g.> $EXEC DB -lib QU -bprm '-filter=none'

To specify a different than default filter, use approach described above. Be sure that specified module exists in $ENV{BLASTFILTER}.

e.g.> $EXEC DB -lib QU -bprm '-filter=nseg'

=head2 MATRIX

To specify a different than default matrix, use -bprm option. Be sure that specified matrix exists in

$ENV{BLASTMAT}

e.g.> $EXEC DB -lib QU -bprm '-filter=nseg -matrix=IDENTITY'

=head2 COMPARISON ON PROTEIN LEVEL

To compare translated nucleotide sequences, use B<$EXEC>'s -bprg option, and set blast program to tblastx.

e.g.> $EXEC DB -lib QU -bprg tblastx

=head2 SEARCH STRANDS

To search only one strand of query use B<$EXEC>'s -bprm option and set -top or -bottom option to blast program.

e.g.> $EXEC DB -lib QU -bprm '-filter=dust -matrix=IDENTITY -bottom'

=head2 PROCESSORS

To limit the number of processors to be used by blast, use B<$EXEC>'s -bprm option and set -p=N (or -cpus=N),
where N - is maximum number of processors to be used.

e.g.> $EXEC DB -hum -bprm '-p=4'

=head2 MULTIPLE DATABASE FILES

To run censor on several files, use unix's find command.

e.g.> /usr/bin/find ./CHR/chr*.fa -exec $EXEC {} -hum \\;

=head1 ENVIRONMENT VARIABLES

=head2 WUBLASTDIR (BLASTDIR)

Wu-blast (NCBI-Blast) installation directory.

As alternative you can edit first lines of script B<$EXEC> to properly set it's internal WUBLASTDIR (BLASTDIR) variable.

=head1 AUTHOR

Written by Oleksiy Kohany

=head1 BUG REPORTS AND COMMENTS

Oleksiy Kohany <kohany\@girinst.org>

=head1 COPYRIGHT

Copyright (C) 2003 Genetic Information Research Institute

=head1 SEE ALSO

B<blastn, blastp, blastx, tblastn, tblastx>

=cut";
   return $usage;
}
1;
