/***************************************************************************
                          atcg.cpp  -  description
                             -------------------
    begin                : Mon Nov 18 2002
    copyright            : (C) 2002 by kohany
    email                : kohany@charon
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "atcg.h"

//--------------------------
  CAtcg::CAtcg() {
    m_dA = m_dT = m_dC = m_dG = 0;;
  }
//--------------------------
  CAtcg::CAtcg(double dA, double dT, double dC, double dG) {
    assign(dA, dT, dC, dG);
  }
//--------------------------
  void CAtcg::assign(double dA, double dT, double dC, double dG) {
    m_dA = dA;
    m_dT = dT;
    m_dC = dC;
    m_dG = dG;
  }
//--------------------------
  CAtcg& CAtcg::operator = (CAtcg& atcg) {
    assign(atcg.getA(), atcg.getT(), atcg.getC(), atcg.getG());
    return *this;
  }
//--------------------------
  void CAtcg::print(FILE *OUTF) {
    fprintf(OUTF, "A = %3.4lf T = %3.4lf C = %3.4lf G = %3.4lf AT = %3.4lf CG = %3.4lf\n",
	    getA(), getT(), getC(), getG(), getAT(), getCG());
  }
//--------------------------
  double CAtcg::getA() { return m_dA; }
//--------------------------
  double CAtcg::getT() { return m_dT; }
//--------------------------
  double CAtcg::getC() { return m_dC; }
//--------------------------
  double CAtcg::getG() { return m_dG; }
//--------------------------
  double CAtcg::getAT() { return m_dA + m_dT; }
//--------------------------
  double CAtcg::getCG() { return m_dC + m_dG; }
//==================================
//--------------------------
  CStats::CStats() {
    reset();
  }
//--------------------------
  void CStats::assign(unsigned long iLength, unsigned long iTotal, unsigned long iA, unsigned long iT, unsigned long iC, unsigned long iG) {
    iLength = iLength;
    iTotal = iTotal;
    iA = iA;
    iT = iT;
    iC = iC;
    iG = iG;
  }
//--------------------------
  void CStats::add(char *p_cLine) {
    while(*p_cLine != '\0') {
      add(*p_cLine++);
    }
  }
//--------------------------
  void CStats::reset() {
    m_iTotal = m_iLength = m_iA = m_iT = m_iC = m_iG = 0;
  }
//--------------------------
  void CStats::add(char cResidue) {
    if(isalpha(cResidue)) {
      m_iTotal++;
      switch (cResidue) {
      case 'a':
      case 'A':
	m_iA++;
	break;
      case 't':
      case 'T':
	m_iT++;
	break;
      case 'c':
      case 'C':
	m_iC++;
	break;
      case 'g':
      case 'G':
	m_iG++;
	break;
      default:
	return;
      }
      m_iLength++;
    }
  }
//--------------------------
  CStats& CStats::operator += (const CStats& stat) {
    return (*this) + stat;
  }
//--------------------------
  CStats& CStats::operator = (const CStats& stat) {
    assign(stat.getLength(), stat.getTotal(), stat.getACount(), stat.getTCount(), stat.getCCount(), stat.getGCount());
    return *this;
  }
//--------------------------
  CStats& CStats::operator + (const CStats& stat) {
    assign(getLength() + stat.getLength(),
	   getTotal() + stat.getTotal(),
	   getACount() + stat.getACount(),
	   getTCount() + stat.getTCount(),
	   getCCount() + stat.getCCount(),
	   getGCount() + stat.getGCount());
	return *this;
  }
//--------------------------
  unsigned long CStats::getACount() const { return m_iA; }
//--------------------------
  unsigned long CStats::getTCount() const { return m_iT; }
//--------------------------
  unsigned long CStats::getCCount() const { return m_iC; }
//--------------------------
  unsigned long CStats::getGCount() const { return m_iG; }
//--------------------------
  double CStats::getARate() const { return (m_iLength == 0) ? 0 : (double(m_iA)) / m_iLength; }
//--------------------------
  double CStats::getTRate() const { return (m_iLength == 0) ? 0 : (double(m_iT)) / m_iLength; }
//--------------------------
  double CStats::getCRate() const { return (m_iLength == 0) ? 0 : (double(m_iC)) / m_iLength; }
//--------------------------
  double CStats::getGRate() const { return (m_iLength == 0) ? 0 : (double(m_iG)) / m_iLength; }
//--------------------------
  unsigned long CStats::getLength() const { return m_iLength; }
//--------------------------
  unsigned long CStats::getTotal() const { return m_iTotal; }
//=======================================
void printMap(FILE *OUTF, CMAP &cmap) {
  CMAP::iterator mptr;
  for(mptr = cmap.begin(); mptr != cmap.end(); mptr++) {
    fprintf(OUTF, "%s ",(*mptr).first.c_str());
    (*mptr).second.print(OUTF);
  }
}
//=======================================
CAtcg& countATCG(FILE *INF, CMAP &cmap) {
  static CAtcg gComp;
  static CStats gStat;
  CStats lStat;
  char line[MAXLINELENGTH];
  char name[MAXNAMESIZE];
  line[0] = '\0';
  gStat.reset();
  while(line[0] == ';' || line[0] == '>' || fgets(line, MAXLINELENGTH, INF) != NULL) {
    switch(line[0]) {
    case '>':
      sscanf(line + 1, "%s", name);
      break;
    case ';':
      while(fgets(line, MAXLINELENGTH, INF) != NULL && line[0] == ';');
      sscanf(line, "%s", name);
      break;
    default:
      continue;
    }
//    fprintf(stderr, "%s\n", name);
    line[0] = '\0';
    lStat.reset();
    while(fgets(line, MAXLINELENGTH, INF) != NULL && line[0] != ';' && line[0] != '>') {
      lStat.add(line);
      gStat.add(line);
    }
    cmap[name].assign(lStat.getARate(), lStat.getTRate(), lStat.getCRate(), lStat.getGRate());
  }
  gComp.assign(gStat.getARate(), gStat.getTRate(), gStat.getCRate(), gStat.getGRate());
  return gComp;
}
//=======================================
CAtcg& countATCG(const SMAP &smap, CMAP &cmap) {
  SMAP::const_iterator mptr;
  static CAtcg gComp;
  CStats gStat;
  for(mptr = smap.begin(); mptr != smap.end(); mptr++) {
    cmap[(*mptr).first].assign((*mptr).second.getARate(), (*mptr).second.getTRate(), (*mptr).second.getCRate(), (*mptr).second.getGRate());
    gStat = gStat + (*mptr).second;
  }
  gComp.assign(gStat.getARate(), gStat.getTRate(), gStat.getCRate(), gStat.getGRate());
  return gComp;
}
//=======================================
CStats& countStat(FILE *INF, SMAP &smap) {
  static CStats gStat;
  SMAP::iterator mptr;
  char line[MAXLINELENGTH];
  char name[MAXNAMESIZE];
  line[0] = '\0';
  gStat.reset();
  while(line[0] == ';' || line[0] == '>' || fgets(line, MAXLINELENGTH, INF) != NULL) {
    switch(line[0]) {
    case '>':
      sscanf(line + 1, "%s", name);
      break;
    case ';':
      while(fgets(line, MAXLINELENGTH, INF) != NULL && line[0] == ';');
      sscanf(line, "%s", name);
      break;
    default:
      continue;
    }
    line[0] = '\0';
    smap[name].reset();
    if((mptr = smap.find(name)) != smap.end()) {
      while(fgets(line, MAXLINELENGTH, INF) != NULL && line[0] != ';' && line[0] != '>') {
	(*mptr).second.add(line);
	gStat.add(line);
      }
    }
    else {
      fprintf(stderr, "Error: smap[%s] was not found\n", name);
    }
  }
  return gStat;
}
//=======================================
void getStat(FILE *INF, SMAP &smap) {
  static CAtcg gComp;
  static CStats gStat;
  char sLine[MAXLINELENGTH];
  char sName[MAXNAMESIZE];
  sLine[0] = 0;
  while(sLine[0] == ';' || sLine[0] == '>' || fgets(sLine, MAXLINELENGTH, INF) != NULL) {
    sName[0] = '\0';
    switch(sLine[0]) {
    case '>':
      sscanf(sLine + 1, "%s", sName);
      break;
    case ';':
      while(fgets(sLine, MAXLINELENGTH, INF) != NULL && sLine[0] == ';');
      sscanf(sLine, "%s", sName);
      break;
    default:
      continue;
    }
//    fprintf(stderr, "%s\n", sName);
    smap[sName].reset();
    while(fgets(sLine, MAXLINELENGTH, INF) != NULL && sLine[0] != ';' && sLine[0] != '>') {
      smap[sName].add(sLine);
    }
//    fprintf(stderr, "%s : %d\n", sName, smap[sName].getLength());
  }
}

